import os
import numpy as np
import multiprocessing as multip

from pydmclab.utils.handy import read_json, write_json, make_sub_for_launcher
from pydmclab.core.query import MPQuery
from pydmclab.core.mag import MagTools
from pydmclab.core.struc import StrucTools
from pydmclab.hpc.launch import LaunchTools
from pydmclab.hpc.submit import SubmitTools
from pydmclab.hpc.analyze import AnalyzeBatch

"""
Basic framework

Initial crystal structures
    a) Sometimes these are entries in Materials Project meeting some criteria:
        - e.g., certain chemical space, chemical formula, E above hull, etc
    b) Other times, we might start with an MP structure then perform some transformations to generate new structures
        - e.g., replace x % of Ru with Ir in Ru_{1-x}Ir_{x}O_2 or extract x Li from Li_{1-x}CoO_2
            - note: now we are disordering structures, so each chemical formula will have >= 1 enumerated structure
                - i.e., Ru_0.5Ir_0.5O_2 has > 1 way to configure Ru/Ir on the cation sublattice
    
    The first two levels of our caculation directories will correspond to information relating to the crystal structure:
        - top_level: the chemical formula (or some surrogate for the chemical formula)
            - e.g., this could be RuO2 in 1(a) or it could be x in 1(b)
        - unique_ID: some unique identifier for a crystal structure belonging to that "top_level"
            - e.g., the Materials Project ID for the 1(a) case; or the index of 
        
    Each unique initial crytsal structure will be used for all directories below top_level/unique_ID
    
    Example tree #1:
    calculating RuO2 and IrO2 ground-state polymorphs
    -calcs
     -RuO2
      -mp-1234
     -IrO2
      -mp-5678
      
    Example tree #2:
    calculating Ru_{2-x}Ir_{x}O4 for x = 0, 1, 2 for 2 ordered structures at each x
    -calcs
     -0
      -0
      -1
     -1
      -0
      -1
     -2
      -0
      -1
         
VASP input settings ("standard")

    Usually we want to compute a group of structures with similar settings so that we can safely compare the resulting energies
    
    - right now, we have two standards:
        - dmc: our group standard (to be evolved collectively)
            - these are best practice settings for our group
        - mp: settings to use for strict comparison to Materials Project data
    - standards define the non-structure VASP input files (INCAR, KPOINTS, POTCAR)
    
    This will be the third level of our calculation directory: top_level/unique_ID/standard
    
    If we were calculating with dmc and mp standards, a part of the tree might look like:
    
    -calcs
     -RuO2
      -mp-1234
       -dmc
       -mp
    
Magnetic configuration (mag)

    We need to define what initial magnetic configuration we want to calculate for each initial structure
    - nm: non-magnetic (no spins)
    - fm: ferromagnetic (all spins > 0)
    - afm_*: antiferromagnetic (spins up and down with sum(spins) = 0)
        - * is a unique identifier for the ordering of spins
            - much like partially occupied / disordered structures have multiple ways to configure ions on the lattice,
                - there are often many ways to configure spins on the lattice while still being AFM (sum(spins) = 0)
        - if we want to do AFM calculations, we should first generate a "MAGMOM" (AFM ordering) file for each unique structure and save that data
        - note: if we are computing AFM, we will also compute FM
        
    This will be the fourth level of our calculation directory: top_level/unique_ID/standard/mag
    
    -calcs
     -RuO2
      -mp-1234
       -dmc
        -afm_0

        
Exchange-correlation functional (xc)

Now, we need to decide what flavor of DFT we want to use. Materials Project uses GGA+U. We will often use METAGGA.

For each xc that we use, it's generally a good idea to run calculations in a sequence:
    - a "loose" calculation that has fast-converging standards but not super accurate
    - a "relax" calculation that will optimize the crystal structure with stricter standards
    - a "static" calculation that will give us a better energy/electronic structure at that optimized geometry
    - for METAGGA calculations, we need to first converge a GGA calculation before we start to optimize the geometry with metagga
    
We don't want to have to submit each of these individually to the queue, so we will "pack" them together:

    So, if we're running gga, that means three calculations will get packed together: gga-loose --> gga-relax --> gga-static
        - if we're running metagga, we'll have 5 calcs: gga-loose --> gga-relax --> gga-static --> metagga-relax --> metagga-static
        - each packing or chain of jobs will need a submission script (i.e., something that gets submitted to the queue)
        
Each one of these "xc-calcs" (e.g., gga-relax) will require a VASP execution, meaning it needs VASP inputs and will generate VASP outputs,
    so each xc-calc gets its own directory, becoming the 5th level of our calculation directory:
    
    top_level/unique_ID/standard/mag/xc-calc
    
    -calcs
     -RuO2
      -mp-1234
       -dmc
        -afm_0
         -gga-loose
         -gga-relax
         -gga-static

"""

"""
Most of this is taken care of automatically in pydmc

1) Use MPQuery to get crystal structures from Materials Project
    - write to a dictionary: query.json
2) [OPTIONAL] Transform those structures to generate new structures using StrucTools
    - write to a dictionary: strucs.json
3) [OPTIONAL] Create AFM orderings ("magmoms") for each structure using MagTools(structure).get_afm_magmoms
    - write to a dictionary: magmoms.json
4) Create a dictionary of "launch directories" using LaunchTools
    - customizable with various _launch_configs
    - launch directories are the directories that hold submission scripts
        - these launch directories are defined by the top_level, unique_ID, standard, and mag
    - this will look like: {top_level/unique_ID/standard/mag : 
                                {'xcs' : [XCs to use for that standard],
                                'magmom' : [magmoms to use for that structure]}
    - write to a dictionary: launch_dirs.json
5) Loop through the launch directories and prepare VASP inputs + submission files and launch each chain of VASP jobs using SubmitTools
    - customizable with _sub_configs, _vasp_configs, and _slurm_configs
    - SubmitTools will figure out which jobs can be submitted together and how to order them for submission
    - It will also prepare each VASP job accordingly
6) Crawl through the launch directories, and analyze every VASP calculation using AnalyzeBatch
    - customizable with _analysis_configs
    - write to a dictionary: results.json
"""

# where is this file
SCRIPTS_DIR = os.getcwd()

# where are my calculations going to live
CALCS_DIR = SCRIPTS_DIR.replace("scripts", "calcs")

# where is my data going to live
DATA_DIR = SCRIPTS_DIR.replace("scripts", "data")

for d in [CALCS_DIR, DATA_DIR]:
    if not os.path.exists(d):
        os.makedirs(d)

# if you need data from MP as a starting point (often the case), you need your API key
API_KEY = "N3KdATtMmcsUL94g"

# lets put a tag on all the files we save
FILE_TAG = CALCS_DIR.split("/")[-2]

# what to query MP for
## e.g., 'MnO2', ['MnO2', 'TiO2'], 'Ca-Ti-O, etc
COMPOSITIONS = "MnO2"

# how to transform MP structures
## e.g., [x/8 for x in range(9)]
## NOTE: you need to modify get_strucs to make this work (hard to generalize)
TRANSFORM_STRUCS = False

# whether or not you want to generate MAGMOMs
## True if you're running AFM, else False
GEN_MAGMOMS = True

# what {standard : [final_xcs]} to calculate
## e.g., {'dmc' : ['metagga', 'ggau']} if you want to run METAGGA + GGA+U at DMC standards
TO_LAUNCH = {"dmc": ["metagga", "ggau"]}

# any configurations related to LaunchTools
## e.g., {'compare_to_mp' : True, 'n_afm_configs' : 4}
LAUNCH_CONFIGS = {"compare_to_mp": True, "n_afm_configs": 1}

# any configurations related to SubmitTools
## usually no need to change anything but n_procs... n_procs will determine how to parallelize over launch_dirs
## NOTE: do not set n_procs = 'all' unless you are running on a compute node (ie not a login node)
SUB_CONFIGS = {"n_procs": 1}

# any configurations related to Slurm
## e.g., {'ntasks' : 16, 'time' : int(24*60)}
SLURM_CONFIGS = {"ntasks": 16, "time": int(24 * 60)}

# any configurations related to VASPSetUp
## e.g., {'lobster_static' : True, 'relax_incar' : {'ENCUT' : 555}}
VASP_CONFIGS = {"lobster_static": True, "relax_incar": {"ENCUT": 555}}

# any configurations related to AnalyzeBatch
## e.g., {'include_meta' : True, 'include_mag' : True, 'n_procs' : 4}
## NOTE: do not set n_procs = 'all' unless you are running on a compute node (ie not a login node)
ANALYSIS_CONFIGS = {"include_meta": True, "include_mag": True, "n_procs": 4}

"""
Don't forget to inspect the arguments to:
    get_query
    get_strucs
    get_magmoms
    get_launch_dirs
    submit_calcs
    get_results

You'll want to customize these depending on your calculations
"""


def get_query(
    comp,
    properties=None,
    criteria=None,
    only_gs=True,
    include_structure=True,
    supercell_structure=False,
    max_Ehull=0.05,
    max_sites_per_structure=65,
    max_strucs_per_cmpd=4,
    savename="query_%s.json" % FILE_TAG,
    remake=False,
):
    """
    Args:
        comp (list or str)
            can either be:
                - a chemical system (str) of elements joined by "-"
                - a chemical formula (str)
            can either be a list of:
                - chemical systems (str) of elements joined by "-"
                - chemical formulas (str)

        properties (list or None)
            list of properties to query
                - if None, then use typical_properties
                - if 'all', then use supported_properties

        criteria (dict or None)
            dictionary of criteria to query
                - if None, then use {}

        only_gs (bool)
            if True, remove non-ground state polymorphs for each unique composition

        include_structure (bool)
            if True, include the structure (as a dictionary) for each entry

        supercell_structure (bool)
            only runs if include_structure = True
            if False, just retrieve the MP structure
            if not False, must be specified as [a,b,c] to make an a x b x c supercell of the MP structure

        max_Ehull (float)
            if not None, remove entries with Ehull_mp > max_Ehull

        max_sites_per_structure (int)
            if not None, remove entries with more than max_sites_per_structure sites

        max_strucs_per_cmpd (int)
            if not None, only retain the lowest energy structures for each composition until you reach max_strucs_per_cmpd

        savename (str) - filename for fjson in DATA_DIR

        remake (bool) - write (True) or just read (False) fjson

    Returns:
        {mpid : {DATA}}
    """

    fjson = os.path.join(DATA_DIR, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    # initialize MPQuery with your API key
    mpq = MPQuery(api_key=API_KEY)

    # get the data from MP
    data = mpq.get_data_for_comp(
        comp=comp,
        properties=properties,
        criteria=criteria,
        only_gs=only_gs,
        include_structure=include_structure,
        supercell_structure=supercell_structure,
        max_Ehull=max_Ehull,
        max_sites_per_structure=max_sites_per_structure,
        max_strucs_per_cmpd=max_strucs_per_cmpd,
    )

    write_json(data, fjson)
    return read_json(fjson)


def check_query(query):
    for mpid in query:
        print("\nmpid: %s" % mpid)
        print("\tcmpd: %s" % query[mpid]["cmpd"])
        print(
            "\tstructure formula: %s"
            % len(StrucTools(query[mpid]["structure"]).formula)
        )


def get_strucs(
    query,
    transform_strucs,
    max_strucs_per_starting_struc=1,
    ox_states=None,
    savename="strucs_%s.json" % FILE_TAG,
    remake=False,
):
    """
    Args:
        query (dict) - {mpid : {DATA}}
        transform_strucs (bool or list):
            - if False, just use the structures from MP
            - if not False:
                - this should be customized to perform whatever transformations you'd like to perform
                - *NOTE*: you should also customize species_mapping in this function
        max_strucs_per_starting_struc (int) - max number of structures to keep for each starting structure
        ox_states (dict) - {element : oxidation state} for transformations
            - if None, use AutoOxidationState algo in pymatgen
        savename (str) - filename for fjson in DATA_DIR
        remake (bool) - write (True) or just read (False) fjson

    Returns:
        if not transform_strucs:
            {formula : {mpid : structure}
        if transform_strucs:
            {formula_identifier : {index_identifier : structure}}
    """

    fjson = os.path.join(DATA_DIR, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    # get all unique chemical formulas in the query
    formulas_in_query = sorted(list(set([query[mpid]["cmpd"] for mpid in query])))

    data = {}
    for formula in formulas_in_query:
        # get all MP IDs in your query having that formula
        mpids = [mpid for mpid in query if query[mpid]["cmpd"] == formula]
        if not transform_strucs:
            # if no transformations, just return the structures from MP
            data[formula] = {mpid: query[mpid]["structure"] for mpid in mpids}
        else:
            # now you need to customize this for whatever transformations you want
            if len(mpids) > 1:
                print("WARNING: %s has %i mpids" % (formula, len(mpids)))
                raise NotImplementedError(
                    "folder structure is not super amenable to this... Reconfiguring recommended"
                )

            # take our 1 MP ID of interest and transform it
            mpid = mpids[0]
            initial_structure = query[mpid]["structure"]
            for x in transform_strucs:
                # make a new "formula" that includes are iterator in the transformation
                key = "_".join([formula, str(x)])
                structools = StrucTools(
                    structure=initial_structure, ox_states=ox_states
                )

                species_mapping = None  # *NOTE*: you should customize this for your desired transformation

                # strucs will have the form {index_identifier : structure}
                strucs = structools.replace_species(
                    species_mapping=species_mapping,
                    n_strucs=max_strucs_per_starting_struc,
                )

                data[key] = strucs

    write_json(data, fjson)
    return read_json(fjson)


def check_strucs(strucs):
    for formula in strucs:
        for ID in strucs[formula]:
            print("\nformula: %s" % formula)
            print("\tID: %s" % ID)
            struc = strucs[formula][ID]
            print("\tstructure formula: %s" % StrucTools(struc).formula)


def get_magmoms(
    strucs,
    max_afm_combos=50,
    treat_as_nm=[],
    savename="magmoms_%s.json" % FILE_TAG,
    remake=False,
):
    """
    Args:
        strucs (dict) - {formula : {ID : structure}}
        max_afm_combos (int): maximum number of AFM spin configurations to generate
        treat_as_nm (list): any normally mag els you'd like to treat as nonmagnetic for AFM enumeration
        savename (str) - filename for fjson in DATA_DIR
        remake (bool) - write (True) or just read (False) fjson

    Returns:
        {formula : {ID : {AFM configuration index : [list of magmoms on each site]}
    """

    fjson = os.path.join(DATA_DIR, savename)
    if not remake and os.path.exists(fjson):
        return read_json(fjson)

    magmoms = {}
    for formula in strucs:
        magmoms[formula] = {}
        for ID in strucs[formula]:
            # for each unique structure, get AFM magmom orderings
            structure = strucs[formula][ID]
            magtools = MagTools(
                structure=structure,
                max_afm_combos=max_afm_combos,
                treat_as_nm=treat_as_nm,
            )
            curr_magmoms = magtools.get_afm_magmoms
            magmoms[formula][ID] = curr_magmoms

    write_json(magmoms, fjson)
    return read_json(fjson)


def check_magmoms(strucs, magmoms):
    for formula in strucs:
        for ID in strucs[formula]:
            structure_formula = StrucTools(strucs[formula][ID]).formula
            n_afm_configs = len(magmoms[formula][ID])
            print("%s: %i AFM configs\n" % (structure_formula, n_afm_configs))


def get_launch_dirs(
    strucs,
    magmoms,
    to_launch,
    user_configs,
    make_launch_dirs=True,
    refresh_configs=True,
    savename="launch_dirs_%s.json" % FILE_TAG,
    remake=False,
):
    """
    Args:
        strucs (dict) - {formula : {ID : structure}}
        magmoms (dict) - {formula : {ID : {AFM configuration index : [list of magmoms on each site]}}
        to_launch (dict) - {standard (str) : [list of final_xcs of interest]}
        user_configs (dict) - optional launch configurations
        make_launch_dirs (bool) - make launch directories (True) or just return launch dict (False)
        refresh_configs (bool) - refresh configs (True) or just use existing configs (False)
        savename (str) - filename for fjson in DATA_DIR
        remake (bool) - write (True) or just read (False) fjson

    Returns:
        {launch_dir (formula/ID/standard/mag) : {'xcs' : [list of final_xcs],
                                                 'magmoms' : [list of magmoms for each site in structure in launch_dir]}}

        also makes launch_dir and populates with POSCAR using strucs if make_dirs=True

    """

    fjson = os.path.join(DATA_DIR, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    all_launch_dirs = {}
    for formula in strucs:
        for ID in strucs[formula]:
            # for each unique structure, generate our launch directories
            structure = strucs[formula][ID]
            if magmoms:
                curr_magmoms = magmoms[formula][ID]
            else:
                curr_magmoms = None
            top_level = formula
            unique_ID = ID

            launch = LaunchTools(
                calcs_dir=CALCS_DIR,
                to_launch=to_launch,
                user_configs=user_configs,
                magmoms=curr_magmoms,
                structure=structure,
                top_level=top_level,
                unique_ID=unique_ID,
                refresh_configs=refresh_configs,
            )

            launch_dirs = launch.launch_dirs(make_dirs=make_launch_dirs)

            for launch_dir in launch_dirs:
                all_launch_dirs[launch_dir] = launch_dirs[launch_dir]

    write_json(all_launch_dirs, fjson)
    return read_json(fjson)


def check_launch_dirs(launch_dirs):
    print("\nanalyzing launch directories")
    for d in launch_dirs:
        print("\nlaunching from %s" % d)
        print("   these final xcs: %s" % launch_dirs[d]["xcs"])


def submit_one_calc(submit_args):

    """
    Prepares VASP inputs, writes submission script, and launches job for one launch_dir

    Args:
        submit_args (dict) should contain:
        {'launch_dir' :
            launch_dir (str) - (formula/ID/standard/mag) to write and launch submission script in,
         'launch_dirs' :
            launch_dirs (dict) - {launch_dir (formula/ID/standard/mag) : {'xcs' : [list of final_xcs], 'magmoms' : [list of magmoms for each site in structure in launch_dir]}},
         'user_configs' :
            user_configs (dict) - optional sub, slurm, or VASP configurations,
         'refresh_configs' :
            refresh_configs (list) - list of which configs to refresh,
         'ready_to_launch':
            ready_to_launch (bool) - write (True) and launch (True) or just write (False) submission scripts (False)
            }

    Returns:
        None

    """
    launch_dir = submit_args["launch_dir"]
    launch_dirs = submit_args["launch_dirs"]
    user_configs = submit_args["user_configs"]
    refresh_configs = submit_args["refresh_configs"]
    ready_to_launch = submit_args["ready_to_launch"]

    # what are our terminal xcs for that launch_dir
    final_xcs = launch_dirs[launch_dir]["xcs"]

    # what magmoms apply to that launch_dir
    magmom = launch_dirs[launch_dir]["magmom"]

    try:
        sub = SubmitTools(
            launch_dir=launch_dir,
            final_xcs=final_xcs,
            magmom=magmom,
            user_configs=user_configs,
            refresh_configs=refresh_configs,
        )

        # prepare VASP directories and write submission script
        sub.write_sub

        # submit submission script to the queue
        if ready_to_launch:
            sub.launch_sub

        success = True
    except TypeError:
        print("\nERROR: %s\n   will submit without multiprocessing" % launch_dir)
        success = False

    return {"launch_dir": launch_dir, "success": success}


def submit_calcs(
    launch_dirs,
    user_configs={},
    refresh_configs=["vasp", "sub", "slurm"],
    ready_to_launch=True,
    n_procs=1,
):
    """
    Prepares VASP inputs, writes submission script, and launches job for all launch_dirs

    Args:
        launch_dirs (dict) - {launch_dir (formula/ID/standard/mag) : {'xcs' : [list of final_xcs], 'magmoms' : [list of magmoms for each site in structure in launch_dir]}}
        user_configs (dict) - optional sub, slurm, or VASP configurations
        refresh_configs (list) - list of which configs to refresh
        ready_to_launch (bool) - write (True) and launch (True) or just write (False) submission scripts (False

    Returns:
        None

    """

    submit_args = {
        "launch_dirs": launch_dirs,
        "user_configs": user_configs,
        "refresh_configs": refresh_configs,
        "ready_to_launch": ready_to_launch,
    }

    if n_procs == 1:
        print("\n\n submitting calculations in serial\n\n")
        for launch_dir in launch_dirs:
            curr_submit_args = submit_args.copy()
            curr_submit_args["launch_dir"] = launch_dir
            submit_one_calc(curr_submit_args)
        return
    elif n_procs == "all":
        n_procs = multip.cpu_count() - 1

    print("\n\n submitting calculations in parallel\n\n")
    print("not refreshing configs for parallel --> causes trouble")
    submit_args["refresh_configs"] = refresh_configs
    list_of_submit_args = []
    for launch_dir in launch_dirs:
        curr_submit_args = submit_args.copy()
        curr_submit_args["launch_dir"] = launch_dir
        list_of_submit_args.append(curr_submit_args)
    pool = multip.Pool(processes=n_procs)
    statuses = pool.map(submit_one_calc, list_of_submit_args)
    pool.close()

    submitted_w_multiprorcessing = [status for status in statuses if status["success"]]
    failed_w_multiprocessing = [status for status in statuses if not status["success"]]

    print(
        "%i/%i calculations submitted with multiprocessing"
        % (len(submitted_w_multiprorcessing), len(statuses))
    )
    for status in failed_w_multiprocessing:
        launch_dir = status["launch_dir"]
        curr_submit_args = submit_args.copy()
        curr_submit_args["launch_dir"] = launch_dir
        submit_one_calc(curr_submit_args)

    return


def get_results(
    launch_dirs,
    user_configs,
    refresh_configs=True,
    savename="results_%s.json" % FILE_TAG,
    remake=False,
):
    """
    Args:
        launch_dirs (dict) - {launch_dir (formula/ID/standard/mag) : {'xcs' : [list of final_xcs], 'magmoms' : [list of magmoms for each site in structure in launch_dir]}}
        user_configs (dict) - optional analysis configurations
        refresh_configs (bool) - refresh configs (True) or just use existing configs (False)
        savename (str) - filename for fjson in DATA_DIR
        remake (bool) - write (True) or just read (False) fjson

    Returns:
        {'formula.ID.standard.mag.xc_calc' : {scraped results from VASP calculation}}
    """

    fjson = os.path.join(DATA_DIR, savename)
    if os.path.exists(fjson) and not remake:
        return read_json(fjson)

    analyzer = AnalyzeBatch(
        launch_dirs, user_configs=user_configs, refresh_configs=refresh_configs
    )

    data = analyzer.results

    write_json(data, fjson)
    return read_json(fjson)


def check_results(results):

    keys_to_check = list(results.keys())

    converged = 0
    for key in keys_to_check:
        top_level, ID, standard, mag, xc_calc = key.split(".")
        data = results[key]
        convergence = results[key]["results"]["convergence"]
        print("\n%s" % key)
        print("convergence = %s" % convergence)
        if convergence:
            converged += 1
            print("E (static) = %.2f" % data["results"]["E_per_at"])
            if ANALYSIS_CONFIGS["include_meta"]:
                print("E (relax) = %.2f" % data["meta"]["E_relax"])
                print("EDIFFG = %i" % data["meta"]["incar"]["EDIFFG"])
                print("1st POTCAR = %s" % data["meta"]["potcar"][0])
            if (
                (mag != "nm")
                and ("include_mag" in ANALYSIS_CONFIGS)
                and (ANALYSIS_CONFIGS["include_mag"])
            ):
                magnetization = data["magnetization"]
                an_el = list(magnetization.keys())[0]
                an_idx = list(magnetization[an_el].keys())[0]
                that_mag = magnetization[an_el][an_idx]["mag"]
                print("mag on %s (%s) = %.2f" % (an_el, str(an_idx), that_mag))
            if ("include_structure" not in ANALYSIS_CONFIGS) or (
                ANALYSIS_CONFIGS["include_structure"]
            ):
                print(data["structure"])

    print("\n\n SUMMARY: %i/%i converged" % (converged, len(keys_to_check)))


def main():

    remake_query = False
    print_query_check = True

    remake_strucs = False
    print_strucs_check = True

    remake_magmoms = False
    print_magmoms_check = True

    remake_launch_dirs = False
    print_launch_dirs_check = True

    remake_subs = True
    ready_to_launch = True

    remake_results = True
    print_results_check = True

    comp = COMPOSITIONS
    query = get_query(comp=comp, remake=remake_query)
    if print_query_check:
        check_query(query)

    transform_strucs = TRANSFORM_STRUCS
    strucs = get_strucs(
        query=query, transform_strucs=transform_strucs, remake=remake_strucs
    )
    if print_strucs_check:
        check_strucs(strucs)

    if GEN_MAGMOMS:
        magmoms = get_magmoms(strucs=strucs, remake=remake_magmoms)
        if print_magmoms_check:
            check_magmoms(strucs=strucs, magmoms=magmoms)
    else:
        magmoms = None

    to_launch = TO_LAUNCH
    launch_configs = LAUNCH_CONFIGS
    launch_dirs = get_launch_dirs(
        strucs=strucs,
        magmoms=magmoms,
        to_launch=to_launch,
        user_configs=launch_configs,
        remake=remake_launch_dirs,
    )
    if print_launch_dirs_check:
        check_launch_dirs(launch_dirs)

    sub_configs = SUB_CONFIGS
    slurm_configs = SLURM_CONFIGS
    vasp_configs = VASP_CONFIGS
    user_sub_configs = {**sub_configs, **slurm_configs, **vasp_configs}
    if remake_subs:
        submit_calcs(
            launch_dirs=launch_dirs,
            user_configs=user_sub_configs,
            ready_to_launch=ready_to_launch,
            n_procs=sub_configs["n_procs"],
        )

    analysis_configs = ANALYSIS_CONFIGS
    results = get_results(
        launch_dirs=launch_dirs, user_configs=analysis_configs, remake=remake_results
    )
    if print_results_check:
        check_results(results)

    return


if __name__ == "__main__":
    main()



from pydmclab.core.comp import CompTools
from pydmclab.utils.handy import eVat_to_kJmol, kJmol_to_eVat


class ReactionEnergy(object):

    """
    *** This is a work in progress ***

    @TODO:
        - write tests/demo
        - incorporate filler
        - incorporate normalization
        - incorporate balance checks

    """

    def __init__(
        self,
        formation_energies,
        reactants,
        products,
        open_to=[],
        norm="atom",
        allowed_filler=["O2", "N2"],
    ):
        """

        Args:
            formation_energies (dict): {formula (str): formation energy (eV/at)}
                - formation energies should account for chemical potentials (e.g., due to partial pressures)
            reactants (dict): {formula (str) : stoichiometry (int)}
            products (dict): {formula (str) : stoichiometry (int)}
            open_to (list): list of elements to be considered "open" in the reaction. Defaults to None.
            norm (str, dict): if 'atom', then calculate reaction energy per atom of products formed
                - otherwise, specify a basis like: {'O' : 3} to normalize per three moles of O in the products formed
        """

        self.formation_energies = formation_energies
        self.reactants = reactants
        self.products = products
        self.open_to = open_to
        self.norm = norm

    @property
    def species(self):
        """
        puts the reactants and products in the same dictionary

        Returns:
            {formula (str) : {'side' : 'left' for reactants, 'right' for products},
                              'amt' : stoichiometry (float) in reaction}}
        """
        species = {}
        reactants, products = self.reactants, self.products
        energies = self.formation_energies
        for r in reactants:
            species[CompTools(r).clean] = {
                "side": "left",
                "amt": reactants[r],
                "Ef": energies[r],
            }
        for p in products:
            species[CompTools(p).clean] = {
                "side": "right",
                "amt": products[p],
                "Ef": energies[p],
            }
        return species

    def check_species_balance(self, species):
        """
        Args:
            species (dict): {formula (str) : {'side' : 'left' for reactants, 'right' for products},
                              'amt' : stoichiometry (float) in reaction}}
        Returns:
            {element (str) : 0 if balanced, else < 0 if more on left, > 0 if more on right}
        """

        involved_elements = [CompTools(formula).els for formula in species]
        involved_elements = sorted(
            list(set([item for sublist in involved_elements for item in sublist]))
        )
        balance = {}
        for el in involved_elements:
            left, right = 0, 0
            for formula in species:
                if el in CompTools(formula).els:
                    if species[formula]["side"] == "left":
                        left += CompTools(formula).els[el] * species[formula]["amt"]
                    elif species[formula]["side"] == "right":
                        right += CompTools(formula).els[el] * species[formula]["amt"]
            balance[el] = left + right

        return left + " --> " + right

    @property
    def E_rxn(self):
        species = self.species
        dE_rxn = 0
        for formula in species:
            if CompTools(formula).n_els == 1:
                continue

            if species[formula]["side"] == "left":
                sign = -1
            elif species[formula]["side"] == "right":
                sign = 1
            else:
                raise ValueError
            coef = species[formula]["amt"]
            Ef = species[formula]["Ef"]
            Ef = eVat_to_kJmol(Ef, formula)
            dE_rxn += sign * coef * Ef

        return dE_rxn



import numpy as np
import math
from itertools import combinations

from pydmclab.data.thermochem import (
    mp2020_compatibility_dmus,
    mus_at_0K,
    mus_at_T,
    mus_from_mp_no_corrections,
    mus_from_bartel2019_npj,
)
from pydmclab.data.features import atomic_masses
from pydmclab.core.comp import CompTools


class ChemPots(object):
    """
    return dictionary of chemical potentials {el : chemical potential (eV/at)} based on user inputs
    """

    def __init__(
        self,
        temperature=0,
        functional="pbe",
        standard="dmc",
        partial_pressures={},  # atm
        diatomics=["H", "N", "O", "F", "Cl"],
        oxide_type="oxide",
        R=8.6173303e-5,  # eV/K
        user_chempots={},
        user_dmus={},
    ):
        """
        Args:
            temperature (int) - temperature in Kelvin
            functional (str) - explicit functional for DFT claculations (don't include +U in name)
            standard (str) - standard for DFT calculations
            partial_pressures (dict) - {el (str) : partial pressure (atm)}
                - adjusts chemical potential of gaseous species based on RTln(p/p0)
            diatomics (list) - list of diatomic elements
                - if el is in diatomics, will use 0.5 * partial pressure effecton mu
            oxide_type (str) - type of oxide
                - this only affects MP Formation energies
                - they use different corrections for oxides, peroxides, and superoxides
            user_chempots (dict) - {el (str) : chemical potential (eV/at)}
                - specifies the chemical potential you want to use for el
                - will override everything
            user_dmus (dict) - {el (str) : delta_mu (eV/at)}
                - specifies the change in chemical potential you want to use for el
                - will override everything except user_chempots
        """
        self.temperature = temperature
        self.functional = functional
        self.standard = standard
        self.partial_pressures = partial_pressures
        self.diatomics = diatomics
        self.oxide_type = oxide_type
        self.R = R
        if standard == "mp":
            mp_dmus = mp2020_compatibility_dmus()
            for el in mp_dmus["anions"]:
                user_dmus[el] = -mp_dmus["anions"][el]
            if functional == "pbeu":
                for el in mp_dmus["U"]:
                    user_dmus[el] = -mp_dmus["U"][el]
            if self.oxide_type == "peroxide":
                user_dmus[el] = -mp_dmus["peroxide"]["O"]
            elif self.oxide_type == "superoxide":
                user_dmus[el] = -mp_dmus["superoxide"]["O"]

        self.user_dmus = user_dmus
        self.user_chempots = user_chempots

    @property
    def chempots(self):
        """
        Returns:
            dictionary of chemical potentials {el : chemical potential (eV/at)} based on user inputs
        """

        if self.temperature == 0:
            if (self.standard == "dmc") or (self.functional in ["scan", "r2scan"]):
                all_mus = mus_at_0K()
                els = sorted(list(all_mus[self.functional].keys()))
                mus = {el: all_mus[self.functional][el]["mu"] for el in els}
            else:
                mus = mus_from_mp_no_corrections()
        else:
            allowed_Ts = list(range(300, 2100, 100))
            if self.temperature not in allowed_Ts:
                raise ValueError("Temperature must be one of %s" % allowed_Ts)
            all_mus = mus_at_T()
            mus = all_mus[str(self.temperature)]

        if self.partial_pressures:
            for el in self.partial_pressures:
                if el in self.diatomics:
                    factor = 1 / 2
                else:
                    factor = 1
                mus[el] += (
                    self.R
                    * self.temperature
                    * factor
                    * np.log(self.partial_pressures[el])
                )
        if self.user_dmus:
            for el in self.user_dmus:
                mus[el] += self.user_dmus[el]
        if self.user_chempots:
            for el in self.user_chempots:
                mus[el] = self.user_chempots[el]

        return mus


class FormationEnthalpy(object):
    """
    For computing formation energies (~equivalently enthalpies) at 0 K

    TO DO:
        - write tests/demo
    """

    def __init__(
        self,
        formula,
        E_DFT,  # eV/at
        chempots,  # from ThermoTools.ChemPots.chempots
    ):

        """
        Args:
            formula (str) - chemical formula
            E_DFT (float) - DFT energy (eV/at)
            chempots (dict) - {el (str) : chemical potential (eV/at)}
                - probably generated using ChemPots.chempots

        """
        self.formula = CompTools(formula).clean
        self.E_DFT = E_DFT
        self.chempots = chempots

    @property
    def weighted_elemental_energies(self):
        """
        Returns:
            weighted elemental energies (eV per formula unit)
        """
        mus = self.chempots
        els_to_amts = CompTools(self.formula).amts
        return np.sum([mus[el] * els_to_amts[el] for el in els_to_amts])

    @property
    def Ef(self):
        """
        Returns:
            formation energy at 0 K (eV/at)
        """
        formula = self.formula
        n_atoms = CompTools(formula).n_atoms
        weighted_elemental_energies = self.weighted_elemental_energies
        E_per_fu = self.E_DFT * n_atoms
        return (1 / n_atoms) * (E_per_fu - weighted_elemental_energies)


class FormationEnergy(object):
    """
    This class is for computing formation energies at T > 0 K

    Automatically uses the Bartel2018 model: https://doi.org/10.1038/s41467-018-06682-4

    TO DO:
        - write tests/demo
        - test ideal mixing entropy?
    """

    def __init__(
        self,
        formula,
        Ef,  # eV/at
        chempots,  # from ThermoTools.ChemPots.chempots
        structure=False,
        atomic_volume=False,
        x_config=None,
        n_config=1,
        include_Svib=True,
        include_Sconfig=False,
    ):

        """
        Args:
            formula (str) - chemical formula
            Ef (float) - DFT formation enthalpy at 0 K (eV/at)
                - or any formation enthalpy at T <= 298 K
            chempots (dict) - {el (str) : chemical potential (eV/at)}
                - probably generated using ChemPots.chempots
            structure (Structure) - pymatgen structure object
                - either structure or atomic_volume needed for vibrational entropy calculation
            atomic_volume (float) - atomic volume (A^3/atom)
                - either structure or atomic_volume needed for vibrational entropy calculation
            override_Ef_0K (float) - formation energy at 0 K (eV/at)
                - if False, compute Ef_0K using FormationEnergy.Ef_0K
            x_config (float) - partial occupancy parameter to compute configurational entropy
                - needed to compute configurational entropy
            n_config (int) - number of systems exhibiting ideal solution behavior
                - this would be one if I have one site that is partially occupied by two ions
                - this would be two if I have two sites that are each partially occupied by two ions
            include_Svib (bool) - whether to include vibrational entropy
            include_Sconfig (bool) - whether to include configurational entropy
        """
        self.formula = CompTools(formula).clean
        self.Ef = Ef
        self.chempots = chempots
        self.structure = structure
        self.atomic_volume = atomic_volume
        self.include_Svib = include_Svib
        self.include_Sconfig = include_Sconfig
        self.x_config = x_config
        self.n_config = n_config

        if include_Svib:
            if not structure and not atomic_volume:
                raise ValueError(
                    "Must provide structure and atomic volume to compute Svib"
                )

        if include_Sconfig:
            if not (x_config and n_config):
                raise ValueError(
                    "Must provide x_config and n_config to compute Sconfig"
                )

    @property
    def weighted_elemental_energies(self):
        """
        Returns:
            weighted elemental energies (eV per formula unit)
        """
        mus = self.chempots
        els_to_amts = CompTools(self.formula).amts
        return np.sum([mus[el] * els_to_amts[el] for el in els_to_amts])

    @property
    def reduced_mass(self):
        """
        Returns weighted reduced mass of composition
            - only needed for G(T) see Chris B Nature Comms 2019
        """
        names = CompTools(self.formula).els
        els_to_amts = CompTools(self.formula).amts
        nums = [els_to_amts[el] for el in names]
        mass_d = atomic_masses()
        num_els = len(names)
        num_atoms = np.sum(nums)
        denom = (num_els - 1) * num_atoms
        if denom <= 0:
            print("descriptor should not be applied to unary compounds (elements)")
            return np.nan
        masses = [mass_d[el] for el in names]
        good_masses = [m for m in masses if not math.isnan(m)]
        if len(good_masses) != len(masses):
            for el in names:
                if math.isnan(mass_d[el]):
                    print("I dont have a mass for %s..." % el)
                    return np.nan
        else:
            pairs = list(combinations(names, 2))
            pair_red_lst = []
            for i in range(len(pairs)):
                first_elem = names.index(pairs[i][0])
                second_elem = names.index(pairs[i][1])
                pair_coeff = nums[first_elem] + nums[second_elem]
                pair_prod = masses[first_elem] * masses[second_elem]
                pair_sum = masses[first_elem] + masses[second_elem]
                pair_red = pair_coeff * pair_prod / pair_sum
                pair_red_lst.append(pair_red)
            return np.sum(pair_red_lst) / denom

    def dGf(self, temperature=0):
        """
        Args:
            temperature (int) - temperature (K)
        Returns:
            formation energy at temperature (eV/at)
                - see Chris B Nature Comms 2019
        """
        T = temperature
        Ef_0K = self.Ef
        if T == 0:
            return Ef_0K
        else:
            if self.include_Svib:
                m = self.reduced_mass
                if self.atomic_volume:
                    V = self.atomic_volume
                elif self.structure:
                    V = self.structure.volume / len(self.structure)
                else:
                    raise ValueError("Need atomic volume or structure to compute G(T)")

                Gd_sisso = (
                    (-2.48e-4 * np.log(V) - 8.94e-5 * m / V) * T
                    + 0.181 * np.log(T)
                    - 0.882
                )
                weighted_elemental_energies = self.weighted_elemental_energies
                G = Ef_0K + Gd_sisso
                n_atoms = CompTools(self.formula).n_atoms

                dGf = (1 / n_atoms) * (G * n_atoms - weighted_elemental_energies)
            else:
                dGf = Ef_0K
                if self.include_Sconfig:
                    x, n = self.x_config, self.n_config
                    kB = 8.617e-5  # eV/K
                    S_config = (
                        -kB * n * (x * np.log(x) + (1 - x) * np.log(1 - x))
                    ) / CompTools(self.formula).n_atoms
                    dGf += -T * S_config
            return dGf


def main():

    mus = ChemPots(functional="r2scan", standard="dmc")

    Ef = FormationEnthalpy(
        formula="IrO2", E_DFT=-0.10729517e04 / 48, chempots=mus.chempots
    ).Ef

    temperature = 2000
    mus = ChemPots(temperature=temperature)

    # return mus, mus
    dGf = FormationEnergy(
        formula="IrO2", Ef=Ef, chempots=mus.chempots, atomic_volume=10.76
    ).dGf(temperature)
    print(Ef)
    print(dGf)

    return mus, fe


if __name__ == "__main__":
    mus, fe = main()



import yaml, os

this_dir, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(this_dir, "data")


def load_vasp_configs():
    with open(os.path.join(DATA_PATH, "_vasp_configs.yaml")) as f:
        return yaml.safe_load(f)


def load_launch_configs():
    with open(os.path.join(DATA_PATH, "_launch_configs.yaml")) as f:
        return yaml.safe_load(f)


def load_slurm_configs():
    with open(os.path.join(DATA_PATH, "_slurm_configs.yaml")) as f:
        return yaml.safe_load(f)


def load_sub_configs():
    with open(os.path.join(DATA_PATH, "_sub_configs.yaml")) as f:
        return yaml.safe_load(f)


def load_partition_configs():
    with open(os.path.join(DATA_PATH, "_partition_configs.yaml")) as f:
        return yaml.safe_load(f)


def load_batch_vasp_analysis_configs():
    with open(os.path.join(DATA_PATH, "_batch_vasp_analysis_configs.yaml")) as f:
        return yaml.safe_load(f)



import os

from pymatgen.core.structure import Structure

from pydmclab.utils.handy import read_yaml, write_yaml, is_calc_valid
from pydmclab.core.mag import MagTools
from pydmclab.data.configs import load_launch_configs
from pydmclab.core.struc import StrucTools

HERE = os.path.dirname(os.path.abspath(__file__))


class LaunchTools(object):
    """
    This is a class to figure out:
        - what launch_dirs need to be created
            - i.e., which directories will house submission scripts
        - what calculation chains need to be run in each launch_dir
    The output is going to be:
        {launch_dir (str) : {'xcs' : [list of final xcs for each chain (str)],
                             'magmom' : [list of magmoms for the structure in that launch_dir (list)],}}
    """

    def __init__(
        self,
        calcs_dir,
        structure,
        top_level,
        unique_ID,
        to_launch,
        magmoms=None,
        user_configs={},
        refresh_configs=True,
        launch_configs_yaml=os.path.join(os.getcwd(), "_launch_configs.yaml"),
    ):
        """
        Args:

            calcs_dir (os.path): directory where calculations will be stored
                - usually if I'm writing a "launch" script to configure and run a bunch of calcs from  a directory: os.getcwd() = */scripts:
                    - then calcs_dir will be os.getcwd().replace('scripts', 'calcs')
                    - I should also probably have a directory to store data called calcs_dir.replace('calcs', 'data')
                    - these are best practices but not strictly enforced in the code anywhere

            structure (Structure): pymatgen structure object
                - usually I want to run a series of calculations for some input structure
                    - this is the input structure

            top_level (str): top level directory
                - could be whatever you want, but there are some guidelines
                    - usually this will be a chemical formula
                        - if I was just running a geometry relaxation on a given chemical formula
                            - let's say LiTiS2)
                                - I would call the top_level = LiTiS2 or even better, Li1S2Ti1 (the CompTools(formula).clean standard)
                - for more complicated calcs,
                    - lets say I'm studying Li_{x}Co10O20 at varying x values between 0 and 10
                        - I might make top_level = LiCoO2

            unique_ID (str): level below top_level
                - could be a material ID in materials project (for standard geometry relaxations, this makes sense)
                - could be x in the LiCoO2 example I described previously
                - it's really up to you, but it must be unique within the top_level directory

            to_launch (dict) :
                {standard (str) : [list of xcs (str)]}
                - e.g., if I want to run gga+u and metagga with dmc standards, to_launch = {'dmc' : ['metagga', 'ggau']}

            magmoms (dict):
                - if you are running AFM calculations
                    - {index of configuration index (int) : magmom (list)} generated using MagTools
                    - best practice is to save this as a json in data_dir so you only call MagTools once
                - if you are not running AFM calculations
                    - can be None or {}

            user_configs (dict):
                - any setting you want to pass that's not default in pydmc/data/data/_launch_configs.yaml
                - launch_configs:
                    compare_to_mp: False # if True, launch will get everything it needs to generate MP-consistent data
                    n_afm_configs: 0 # how many AFM configurations to run
                    override_mag: False # could be ['nm'] if you only want to run nonmagnetic, won't check for whether structure is mag or not mag, it will just do as you say


            refresh_configs (bool) - if True, will copy pydmc baseline configs to your local directory
                - this is useful if you've made changes to the configs files in the directory you're working in and want to start over

            launch_configs_yaml (os.pathLike) - path to yaml file containing launch configs
                - there's usually no reason to change this
                - this holds some default configs for LaunchTools
                - can always be changed with user_configs

        Returns:
            configs (dotdict): dictionary of all configs and arguments to LaunchTools
        """

        # make our calcs_dir if it doesn't exist (this will hold all the launch_dirs)
        if not os.path.exists(calcs_dir):
            os.mkdir(calcs_dir)

        # make our local launch_configs file if it doesn't exist
        if not os.path.exists(launch_configs_yaml) or refresh_configs:
            _launch_configs = load_launch_configs()
            write_yaml(_launch_configs, launch_configs_yaml)

        # initialize our baseline launch_configs
        _launch_configs = read_yaml(launch_configs_yaml)

        # update our baseline launch_configs with user_configs
        configs = {**_launch_configs, **user_configs}

        # make structure a dict() for easier handling
        if not isinstance(structure, dict):
            structure = structure.as_dict()

        # check to make sure we have magmoms if we're running AFM calcs
        if configs["n_afm_configs"] > 0:
            if MagTools(structure).could_be_afm:
                if not magmoms:
                    raise ValueError(
                        "You are running afm calculations but provided no magmoms, generate these first, then pass to LaunchTools"
                    )

        # include gga+u calcs w/ mp standards if we're comparing to MP
        if configs["compare_to_mp"]:
            to_launch["mp"] = ["ggau"]

        # add the required arguments to our configs file
        configs["top_level"] = top_level
        configs["unique_ID"] = unique_ID
        configs["calcs_dir"] = calcs_dir
        configs["to_launch"] = to_launch

        # store our magmoms and structure
        self.magmoms = magmoms
        self.structure = structure

        # make a copy of our configs to prevent unwanted changes
        self.configs = configs.copy()

    @property
    def valid_mags(self):
        """
        Returns:
            list of magnetic configuration names that make sense to run based on the inputs

        e.g.,
            - if we have a nonmagnetic system, this might be ['nm']
            - if we set n_afm_configs = 100, but our magmoms only has 3 configs, then this will just hold ['fm', 'afm_0', 'afm_1', 'afm_2']

        these are the set of "mags" that can be run given our inputs

        Note:
            - configs['override_mag'] will force that we use configs['override_mag'] as our mag

        """
        # copy our configs
        configs = self.configs.copy()

        # return override_mag if we set it
        if configs["override_mag"]:
            return configs["override_mag"]

        structure = self.structure

        # if we're not magnetic, return nm
        if not MagTools(structure).could_be_magnetic:
            return ["nm"]

        # if we can't be AFM or we didn't ask for AFM, but we are magnetic, return fm
        if not MagTools(structure).could_be_afm or not configs["n_afm_configs"]:
            return ["fm"]

        # figure out the max AFM index we can run based on what we asked for

        max_desired_afm_idx = configs["n_afm_configs"] - 1

        magmoms = self.magmoms

        configs_in_magmoms = list(magmoms.keys())
        configs_in_magmoms = sorted([int(i) for i in configs_in_magmoms])
        max_available_afm_idx = max(configs_in_magmoms)

        max_afm_idx = min(max_desired_afm_idx, max_available_afm_idx)

        afm_indices = ["afm_%s" % str(i) for i in range(max_afm_idx + 1)]

        return ["fm"] + afm_indices

    def launch_dirs(self, make_dirs=True):
        """
        Args:
            make_dirs (bool) - if True, make the launch_dir and populate it with a POSCAR
        Returns:
            a dictionary of:
                {launch_dir (str) : {'xcs': [list of final_xcs to submit],
                                     'magmom' : [list of magmoms for the structure in launch_dir]}}

        Returns the minimal list of directories that will house submission files (each of which launch a chain of calcs)
            - note a chain of calcs must have the same structure and magnetic information, otherwise, there's no reason to chain them

        These launch_dirs have a very prescribed structure:
            calcs_dir / top_level / unique_ID / standard / mag

            e.g.,
                - ../calcs/Nd2O7Ru2/mp-19930/dmc/fm
                - ../calcs/2/3/dmc/afm_4
                    - (if (2) was a unique compositional indicator and (3) was a unique structural indicator)
        """
        structure = self.structure

        # make a copy of our configs to prevent unwanted changes
        configs = self.configs.copy()

        # the list of mags we can run
        mags = self.valid_mags

        magmoms = self.magmoms

        # final_xcs we want to run for each standard
        to_launch = configs["to_launch"]

        # level0 houses all our launch_dirs
        level0 = configs["calcs_dir"]

        # level1 describes the composition
        level1 = configs["top_level"]

        # level2 describes the structure
        level2 = configs["unique_ID"]

        launch_dirs = {}
        for standard in to_launch:
            # for each standard we asked for, use that as level3
            level3 = standard

            # we asked for certain xcs at each standard, hold them here
            xcs = to_launch[standard]

            for mag in mags:
                # for each mag we can run, use that as level4
                level4 = mag
                magmom = None
                if "afm" in mag:
                    # grab the magmom if our calc is AFM
                    idx = mag.split("_")[1]
                    if str(idx) in magmoms:
                        magmom = magmoms[str(idx)]
                    elif int(idx) in magmoms:
                        magmom = magmoms[int(idx)]

                # our launch_dir is now defined
                launch_dir = os.path.join(level0, level1, level2, level3, level4)

                # save the final_xcs we want to submit in this launch_dir
                # SubmitTools will make 1 submission script for each final_xc
                # save the magmom as well. VASPSetUp will need that to set the INCARs for all calcs in this launch_dir
                launch_dirs[launch_dir] = {"xcs": xcs, "magmom": magmom}

                # if make_dirs, make the launch_dir and put a POSCAR in there
                if make_dirs:
                    if not os.path.exists(launch_dir):
                        os.makedirs(launch_dir)
                    fposcar = os.path.join(launch_dir, "POSCAR")
                    if not os.path.exists(fposcar):
                        struc = Structure.from_dict(structure)
                        if configs["perturb_launch_poscar"]:
                            initial_structure = struc.copy()
                            if isinstance(configs["perturb_launch_poscar"], bool):
                                perturbation = 0.05
                            else:
                                perturbation = configs["perturb_launch_poscar"]
                            perturbed_structure = StrucTools(initial_structure).perturb(
                                perturbation
                            )
                            perturbed_structure.to(fmt="poscar", filename=fposcar)
                        else:
                            struc.to(fmt="poscar", filename=fposcar)

        return launch_dirs



import matplotlib as mpl
from pydmclab.data.plotting_configs import get_color_palettes


def get_colors(palette):
    """

    returns rgb colors that are nicer than matplotlibs defaults

    Args:
        palette (str):
            'tab10' : tableau 10 colors
            'paired' : "paired" light and dark colors
            'set2' : pastel-y colors
            'dark2' : dark pastel-y colors

        For reference, see: https://matplotlib.org/stable/_images/sphx_glr_colormaps_006.png


    Returns:
        {color (str) : rgb (tuple)}

        so, to use this you could do:
            from pydmc.utils.plotting import get_colors
            my_colors = get_colors('tab10')
            ax = plt.scatter(x, y, color=my_colors['blue'])
    """
    colors = get_color_palettes()[palette]
    return colors


def set_rc_params():
    """
    Args:

    Returns:
        dictionary of settings for mpl.rcParams
    """
    params = {
        "axes.linewidth": 1.5,
        "axes.unicode_minus": False,
        "figure.dpi": 300,
        "font.size": 20,
        "legend.frameon": False,
        "legend.handletextpad": 0.4,
        "legend.handlelength": 1,
        "legend.fontsize": 12,
        "mathtext.default": "regular",
        "savefig.bbox": "tight",
        "xtick.labelsize": 20,
        "ytick.labelsize": 20,
        "xtick.major.size": 6,
        "ytick.major.size": 6,
        "xtick.major.width": 1.5,
        "ytick.major.width": 1.5,
        "xtick.top": True,
        "ytick.right": True,
        "axes.edgecolor": "black",
        "figure.figsize": [6, 4],
    }
    for p in params:
        mpl.rcParams[p] = params[p]
    return params



from pydmclab.core.comp import CompTools

"""
This file currently tests the following classes in CompTools
    - CompTools

TO-DO:
    - convert into formal "tests"
"""


def print_initial_final(initial, final):
    """
    Args:
        initial (str) - initial formula
        final (str) - final formula

    Returns:
        prints the initial and final formulas
    """

    print("\ninitial: %s --> %s (final)" % (initial, final))


def demo_basic_formula_manipulations():
    """
    prints initial formulas and CompTools(initial).clean formulas
        - these should be:
            - alphabetized
            - strings with no spaces
            - integer amounts for each element
            - reduced (i.e., cannot be simplified to smaller integers)

    """
    print("\n")
    initial_formulas = ["TiO2", "NaV2(PO4)3", "Ga20As40", "AlO1.5"]

    for initial in initial_formulas:
        print_initial_final(initial, CompTools(initial).clean)

    return


def demo_amts():
    """
    prints the extracted integer amount of each element in the formula
    """
    print("\n")
    initial = "NaV2.5(PO4)3"

    ct = CompTools(initial)
    clean_formula = ct.clean
    print("\nclean formula: %s\n" % clean_formula)

    amts_dict = ct.amts
    for el in amts_dict:
        print("%i %s in %s" % (amts_dict[el], el, clean_formula))


def demo_mol_frac_and_stoich():
    """
    prints the extracted integer amount and fractional amount of each element in the formula
    """
    print("\n")
    initial = "Cs2AgInCl6"

    ct = CompTools(initial)
    clean_formula = ct.clean
    print("\nclean formula: %s\n" % clean_formula)

    els = ct.els
    for el in els:
        print(
            "%i %s in %s; fractional amount = %.2f"
            % (ct.stoich(el), el, clean_formula, ct.mol_frac(el))
        )


def demo_diatomics():
    """
    ensures that diatomic elements get simplifed to O1, N1, etc
        - this is desired behavior for ThermoTools
        - might need to modify CompTools.label_for_plot to make these diatomic
    """
    print("\n")
    initial_formulas = ["O2", "H2", "Cl2", "Br2", "N2"]
    for initial in initial_formulas:
        print(CompTools(initial).clean)


def demo_plot_labels():
    """
    converts formula strings into strings with subscripts for plotting purposes
    """
    initial_formulas = ["LiMn2O4", "MgCr2S4", "Si1.6C1.6", "MgAlO3.5"]

    print("\n")
    el_order = ("Li", "Si", "Mg", "Al", "Cr", "Mn", "C", "O")
    for initial in initial_formulas:
        ct = CompTools(initial)
        clean_formula = ct.clean
        print("%s = %s" % (clean_formula, ct.label_for_plot(el_order=el_order)))


def demo_simple_comp():
    formula = "NaV2(PO4)3"
    ct_obj = CompTools(formula)
    return ct_obj


def main():
    # return demo_simple_comp()
    demo_basic_formula_manipulations()
    demo_amts()
    demo_mol_frac_and_stoich()
    demo_diatomics()
    demo_plot_labels()
    return


if __name__ == "__main__":
    a = main()



import os
import numpy as np
import multiprocessing as multip
import json

from pymatgen.io.vasp.outputs import Vasprun, Outcar, Eigenval
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Kpoints, Incar
from pymatgen.io.lobster.outputs import Doscar, Cohpcar, Charge, MadelungEnergies

from pydmclab.core.struc import StrucTools, SiteTools
from pydmclab.utils.handy import read_json, write_json, read_yaml, write_yaml
from pydmclab.data.configs import load_batch_vasp_analysis_configs


class VASPOutputs(object):
    def __init__(self, calc_dir):
        """

        Args:
            calc_dir (str): path to where you ran VASP
        """

        self.calc_dir = calc_dir

    @property
    def vasprun(self):
        """
        Returns Vasprun object from vasprun.xml in calc_dir
        """
        fvasprun = os.path.join(self.calc_dir, "vasprun.xml")
        if not os.path.exists(fvasprun):
            return None

        try:
            vr = Vasprun(os.path.join(self.calc_dir, "vasprun.xml"))
            return vr
        except:
            return None

    @property
    def poscar(self):
        """
        Returns Structure object from POSCAR in calc_dir
        """
        fposcar = os.path.join(self.calc_dir, "POSCAR")
        if not os.path.exists(fposcar):
            return None

        try:
            return Structure.from_file(os.path.join(self.calc_dir, "POSCAR"))
        except:
            return None

    @property
    def incar(self):
        """
        Returns dict of VASP input settings from INCAR
        """
        fincar = os.path.join(self.calc_dir, "INCAR")
        if os.path.exists(fincar):
            return Incar.from_file(os.path.join(self.calc_dir, "INCAR"))
        else:
            return {}

    @property
    def all_input_settings(self):
        """
        Returns dict of VASP input settings from vasprun (ie what VASP actually used)
        """
        vr = self.vasprun
        if vr:
            return vr.parameters
        return {}

    @property
    def kpoints(self):
        """
        Returns Kpoints object from KPOINTS in calc_dir
        """
        fkpoints = os.path.join(self.calc_dir, "KPOINTS")

        if not os.path.exists(fkpoints):
            return None
        try:
            return Kpoints.from_file(fkpoints)
        except:
            return None

    @property
    def actual_kpoints(self):
        """
        Returns actual kpoints that were used (list of [a,b,c] for each kpoint)
        """
        vr = self.vasprun
        if vr:
            return vr.actual_kpoints
        else:
            return None

    @property
    def potcar(self):
        """
        Returns {el : {'name' : POTCAR name, 'pp' : pseudopotential name, 'date' : date of POTCAR, 'nval' : number of valence electrons}}}
        """
        fpot = os.path.join(self.calc_dir, "POTCAR")
        if not os.path.exists(fpot):
            return None
        out = {}
        with open(fpot) as f:
            for line in f:
                if "VRHFIN" in line:
                    el = line.split("=")[1].split(":")[0].strip()
                if "TITEL" in line:
                    tmp_dict = {}
                    line = line.split("=")[1].split(" ")
                    tmp_dict["name"] = line[1]
                    tmp_dict["pp"] = line[2]
                    tmp_dict["date"] = line[3][:-1]
                if "ZVAL" in line:
                    line = line.split(";")[1].split("=")[1].split("mass")[0]
                    tmp_dict["nval"] = int(
                        float("".join([val for val in line if val != " "]))
                    )
                    out[el] = tmp_dict
        return out

    @property
    def contcar(self):
        """
        Returns Structure object from CONTCAR in calc_dir
        """
        fcontcar = os.path.join(self.calc_dir, "CONTCAR")
        if not os.path.exists(fcontcar):
            return None
        try:
            return Structure.from_file(os.path.join(self.calc_dir, "CONTCAR"))
        except:
            return None

    @property
    def outcar(self):
        """
        Returns Outcar object from OUTCAR in calc_dir
        """
        foutcar = os.path.join(self.calc_dir, "OUTCAR")
        if not os.path.exists(foutcar):
            return None

        try:
            oc = Outcar(os.path.join(self.calc_dir, "OUTCAR"))
            return oc
        except:
            return None

    @property
    def eigenval(self):
        """
        Returns Eigenval object from EIGENVAL in calc_dir
        """
        feigenval = os.path.join(self.calc_dir, "EIGENVAL")
        if not os.path.exists(feigenval):
            return None

        try:
            ev = Eigenval(os.path.join(self.calc_dir, "EIGENVAL"))
            return ev
        except:
            return None

    def doscar(self, fdoscar="DOSCAR.lobster"):
        """
        fdoscar (str) - 'DOSCAR' or 'DOSCAR.lobster'

        Returns:
            Doscar object from DOSCAR in calc_dir
        """
        fdoscar = os.path.join(self.calc_dir, fdoscar)
        if not os.path.exists(fdoscar):
            return None

        try:
            dos = Doscar(
                doscar=fdoscar, structure_file=os.path.join(self.calc_dir, "CONTCAR")
            )
        except:
            return None

        return dos

    def cohpcar(self, are_coops=False, are_cobis=False):
        """
        Args:
            are_coops (bool) : if True, use COOPCAR
            are_cobis (bool) : if True, use COBI?
        """
        if are_coops:
            flobster = os.path.join(self.calc_dir, "COOPCAR.lobster")
        elif are_cobis:
            flobster = os.path.join(self.calc_dir, "COBICAR.lobster")
        else:
            flobster = os.path.join(self.calc_dir, "COHPCAR.lobster")

        if not os.path.exists(flobster):
            return None

        try:
            cohp = Cohpcar(are_coops=are_coops, are_cobis=are_cobis, filename=flobster)
        except:
            return None

        return cohp

    @property
    def lobsterin(self):
        """
        helps set up Lobsterin file

        Returns:
            {el (str) : {'orbs' : [orbital names], 'basis' : basis name}}
        """
        s_orbs = ["s"]
        p_orbs = ["p_x", "p_y", "p_z"]
        d_orbs = ["d_xy", "d_yz", "d_z^2", "d_xz", "d_x^2-y^2"]
        f_orbs = [
            "f_y(3x^2-y^2)",
            "f_xyz",
            "f_yz^2",
            "f_z^3",
            "f_xz^2",
            "f_z(x^2-y^2)",
            "f_x(x^2-3y^2)",
        ]
        all_orbitals = {"s": s_orbs, "p": p_orbs, "d": d_orbs, "f": f_orbs}

        basis_functions = {}
        with open(os.path.join(self.calc_dir, "lobsterin")) as f:
            for line in f:
                if "basisfunctions" in line:
                    line = line[:-1].split(" ")
                    el = line[1]
                    basis = line[2:-1]
                    basis_functions[el] = basis

        # print(data)
        orbs = {}
        for el in basis_functions:
            orbs[el] = []
            for basis in basis_functions[el]:
                number = basis[0]
                letter = basis[1]
                # print(number)
                # print(letter)
                # print('\n')
                orbitals = [number + v for v in all_orbitals[letter]]
                orbs[el] += orbitals

        data = {el: {"orbs": orbs[el], "basis": basis_functions[el]} for el in orbs}

        return data


class AnalyzeVASP(object):
    """
    Analyze the results of one VASP calculation
    """

    def __init__(self, calc_dir, calc=None):
        """
        Args:
            calc_dir (os.PathLike) - path to directory containing VASP calculation
            calc (str) = what kind of calc was done in calc_dir
                - if None, infer from calc_dir
                - could also be in ['loose', 'static', 'relax']

        Returns:
            calc_dir, calc
            outputs (VASPOutputs(calc_dir))

        """

        self.calc_dir = calc_dir
        if not calc:
            self.calc = os.path.split(calc_dir)[-1].split("-")[1]
        else:
            self.calc = calc

        self.outputs = VASPOutputs(calc_dir)

    @property
    def is_converged(self):
        """
        Returns True if VASP calculation is converged, else False
        """
        vr = self.outputs.vasprun
        if vr:
            if self.calc == "static":
                return vr.converged_electronic
            else:
                return vr.converged
        else:
            return False

    @property
    def E_per_at(self):
        """
        Returns energy per atom (eV/atom) from vasprun.xml or None if calc not converged
        """
        if self.is_converged:
            vr = self.outputs.vasprun
            nsites = self.nsites

            if nsites and vr.final_energy:
                return vr.final_energy / nsites
            return None
        else:
            return None

    @property
    def nsites(self):
        """
        Returns number of sites in POSCAR
        """
        poscar = self.outputs.poscar
        if poscar:
            return len(poscar)
        else:
            return None

    @property
    def nbands(self):
        """
        Returns number of bands from EIGENVAL
        """
        ev = self.outputs.eigenval
        if ev:
            return ev.nbands
        else:
            return None

    @property
    def sites_to_els(self):
        """
        Returns {site index (int) : element (str) for every site in structure}
        """
        contcar = self.outputs.contcar
        if not contcar:
            return None
        return {idx: SiteTools(contcar, idx).el for idx in range(len(contcar))}

    @property
    def magnetization(self):
        """
        Returns {element (str) :
                    {site index (int) :
                        total magnetization on site (float)}}
        """
        oc = self.outputs.outcar
        if not oc:
            return {}
        mag = list(oc.magnetization)
        if not mag:
            return {}
        sites_to_els = self.sites_to_els
        els = sorted(list(set(sites_to_els.values())))
        return {
            el: {
                str(idx): mag[idx]["tot"]
                for idx in sorted([i for i in sites_to_els if sites_to_els[i] == el])
            }
            for el in els
        }

    @property
    def gap_properties(self):
        """
        Returns {'bandgap' : bandgap (eV),
                 'is_direct' : True if gap is direct else False,
                 'cbm' : cbm (eV),
                 'vbm' : vbm (eV)}
        """
        vr = self.outputs.vasprun
        if vr:
            props = vr.eigenvalue_band_properties
            return {
                "bandgap": props[0],
                "cbm": props[1],
                "vbm": props[2],
                "is_direct": props[3],
            }
        else:
            return None

    @property
    def els_to_orbs(self):
        """
        Returns:
            {element (str) : [list of orbitals (str) considered in calculation]}

        """
        lobsterin = self.outputs.lobsterin
        return {el: lobsterin[el]["orbs"] for el in lobsterin}

    @property
    def sites_to_orbs(self):
        """
        Returns:
            {site index (int) : [list of orbitals (str) considered in calculation]}
        """
        sites_to_els = self.sites_to_els
        els_to_orbs = self.els_to_orbs
        out = {}
        for site in sites_to_els:
            el = sites_to_els[site]
            orbs = els_to_orbs[el]
            out[site] = orbs
        return out

    def pdos(self, fjson=None, remake=False):
        """
        @TODO: add demo/test

        Returns complex dict of projected DOS data
            - uses DOSCAR.lobster as of now (must run LOBSTER first)

        Args:
            fjson (os.PathLike) - path to json file to save pdos data to
            remake (bool) - if True, remake pdos data even if fjson exists

        Returns:
            - dictionary (let's call it d)
            - list(d.keys()) = ['E'] + [list of elements (str)]
                - e.g., if I calculate Ca2Ru2O7, the keys will be ['E', 'Ca', 'Ru', 'O']
            - d['E'] = 1d array of energies corresponding with DOS
            - d[el] =
                {site index in CONTCAR (str) :
                    {orbital (str) (e.g., '2p_x') :
                        spin (str) (e.g., '1' (spin up) or '-1' (spin down) :
                            DOS (1d array)}}
                - elaborating
                    - d[el].keys() would be all the sites where that element appears in the structure ['4', '5', '6', ...]
                    - d[el][site].keys() would be the orbitals for all valence electrons in the calculation ['4d_xy', '4d_xz', '4d_yz', ...]
                    - d[el][site][orb].keys() would be the spins ['1', '-1']

            - so if I wanted to plot the spin-up 4d_xy DOS for Ru on site 8, it would be:
                energies = d['E']
                dos = d['Ru']['8']['4d_xy']['1']
                plt.plot(dos, energies)

        """

        if not fjson:
            fjson = os.path.join(self.calc_dir, "pdos.json")
        if os.path.exists(fjson) and not remake:
            try:
                return read_json(fjson)
            except json.decoder.JSONDecodeError:
                pass

        doscar = self.outputs.doscar()
        if not doscar:
            return None

        complete_dos = doscar.completedos

        sites_to_orbs = self.sites_to_orbs
        sites_to_els = self.sites_to_els
        structure = self.outputs.contcar
        out = {}
        for site_idx in sites_to_orbs:
            site = structure[site_idx]
            el = sites_to_els[site_idx]
            orbitals = sites_to_orbs[site_idx]
            if el not in out:
                out[el] = {str(site_idx): {}}
            else:
                out[el][str(site_idx)] = {}
            for orbital in orbitals:
                out[el][str(site_idx)][orbital] = {}
                dos = complete_dos.get_site_orbital_dos(site, orbital).as_dict()
                energies = dos["energies"]
                for spin in dos["densities"]:
                    out[el][str(site_idx)][orbital][spin] = dos["densities"][spin]
        out["E"] = energies
        write_json(out, fjson)
        return read_json(fjson)

    def tdos(self, pdos=None, fjson=None, remake=False):
        """
        Returns more compact dict than pdos
            - uses DOSCAR.lobster as of now (must run LOBSTER first)

        Args:
            pdos (dict or None) - if you don't pass the pdos dictionary, it will make it before working on tdos
            fjson (os.PathLike) - path to json file to save pdos data to
            remake (bool) - if True, remake pdos data even if fjson exists

        Returns:
            - dictionary (let's call it d)
            - list(d.keys()) = ['E', 'total', 'up', 'down'] + [list of elements (str)]
            - d['E'] = 1d array of energies corresponding with DOS
            - d[el] = 1d array of DOS for that element (sums all sites, orbitals, and spins)
            - d['total'] = 1d array of DOS for structure (sums all elements, sites, orbitals, spins)
            - d['up'] or d['down']:
                - keys are ['total'] + [list of elements (str)]
                - d['up']['total'] = 1d array of spin-up DOS for structure
                - d['down'][el] = 1d array of spin-down DOS for that element
                - etc

            - so if I wanted to plot the total DOS for my structure and separate spin up (+ DOS) and spin down (- DOS)

                energies = d['E']
                dos_up = d['up']['total']
                dos_down = d['down']['total']
                plt.plot(dos_up, energies)
                plt.plot(-1*dos_down, energies)

        @TODO: add demo/test (reasonably well tested now)
        @TODO: work on generic plotting
        """
        if not fjson:
            fjson = os.path.join(self.calc_dir, "tdos.json")
        if os.path.exists(fjson) and not remake:
            try:
                return read_json(fjson)
            except json.decoder.JSONDecodeError:
                pass
        if not pdos:
            pdos = self.pdos()
        if not pdos:
            return None
        out = {"up": {}, "down": {}}
        energies = pdos["E"]
        for el in pdos:
            if el == "E":
                continue
            out[el] = np.zeros(len(energies))
            out["up"][el] = np.zeros(len(energies))
            out["down"][el] = np.zeros(len(energies))
            for site in pdos[el]:
                for orb in pdos[el][site]:
                    for spin in pdos[el][site][orb]:
                        if int(spin) == 1:
                            out["up"][el] += np.array(pdos[el][site][orb][spin])
                        elif int(spin) == -1:
                            out["down"][el] += np.array(pdos[el][site][orb][spin])
                        out[el] += np.array(pdos[el][site][orb][spin])
        out["total"] = np.zeros(len(energies))
        for el in out:
            if el in ["up", "down", "total"]:
                continue
            out["total"] += np.array(out[el])
        for spin in ["up", "down"]:
            out[spin]["total"] = np.zeros(len(energies))
            for el in out[spin]:
                if el != "total":
                    out[spin]["total"] += np.array(out[spin][el])
        out["E"] = energies
        for k in out:
            if k not in ["up", "down"]:
                out[k] = list(out[k])
        for spin in ["up", "down"]:
            for k in out[spin]:
                out[spin][k] = list(out[spin][k])
        write_json(out, fjson)
        return read_json(fjson)

    def pcohp(self, fjson=None, remake=False, are_coops=False, are_cobis=False):
        """
        @TODO: add demo/test

        Returns complex dict of projected COHP, COOP, or COBI data
            - must run LOBSTER first
            - note: COHP, COOP, and COBI have similar format so they get processed the same way. If you want all three, you might do:
                for bonding in ["cohp", "coop", "cobi"]:
                    cohp_data = self.pcohp(are_coops=False, are_cobis=False)
                    coop_data = self.pcohp(are_coops=True, are_cobis=False)
                    cobi_data = self.pcohp(are_coops=False, are_cobis=True)

        Args:
            fjson (os.PathLike) - path to json file to save pdos data to
            remake (bool) - if True, remake pchohp data even if fjson exists
            are_coops (bool) - True if you want COOP analysis
            are_cobis (bool) - True if you want COBI analysis
                - both False = COHP analysis

        Returns:
            - dictionary (let's call it d)
            - list(d.keys()) = ['E'] + [el1-el2 (str) for all pairs of elements in structure that might be bonding]
                - note: el1-el2 is sorted alphabetically
                - e.g., the keys for Ca2Ru2O7 might be: ['Ca-Ca', 'Ca-Ru', 'Ca-O', 'Ru-Ru', 'O-Ru', 'O-O', 'E']
            - d['E'] = 1d array of energies corresponding with DOS
            - for each "bond type" (el1-el2), we have particular bonds:
                - d[el1-el2].keys() might be ['16-8', '17-8', '18-8', '16-9', ...]
                    - indicating specific between el1 on site 16 and el2 on site 8, etc
                - for each specific bond, we have three keys:
                    d[el1-el2][site1-site].keys() = ['cohp', 'icohp', 'length']
                    - 'length' returns simply the bond length in Angstrom (float)
                    - 'cohp' and 'icohp' have the same structure
                        - 'cohp' refers to the absolute COHP/COOP/COBI and 'icohp' refers to the integrated COHP/COOP/COBI (don't remember exactly how this integral is done but I remember not loving it...)
                        - d[el1-el2][site1-site]['cohp'].keys() = ['1', '-1', 'total']
                            - '1' and '-1' refer to spin up and spin down, respectively
                            - 'total' means summing over these two spins
                            - d[el1-el2][site1-site]['cohp']['1' or '-1' or 'total'] gives a 1d array of populations

            - so if I wanted to plot the total COHP for the interaction between Ru on site 8 and O on site 16, I would do:
                energies = d['E']
                cohp = d['O-Ru']['16-8']['cohp']['total']
                plt.plot(cohp, energies)

        """

        if are_coops:
            savename = "pcoop.json"
        elif are_cobis:
            savename = "pcobi.json"
        else:
            savename = "pcohp.json"

        if not fjson:
            fjson = os.path.join(self.calc_dir, savename)
        if os.path.exists(fjson) and not remake:
            try:
                return read_json(fjson)
            except json.decoder.JSONDecodeError:
                pass

        cohpcar = self.outputs.cohpcar(are_coops=are_coops, are_cobis=are_cobis)
        if not cohpcar:
            return None

        energies = cohpcar.energies

        cohp = cohpcar.cohp_data

        sites_to_els = self.sites_to_els

        out = {}
        for bond_idx in cohp:
            if bond_idx == "average":
                continue
            sites = cohp[bond_idx]["sites"]
            els = [sites_to_els[site] for site in sites]
            if sorted(els) == els:
                sorted_sites = sites
            else:
                sorted_sites = (sites[1], sites[0])
            el_tag = "-".join(sorted(els))
            site_tag = "-".join([str(site) for site in sorted_sites])
            bond_length = cohp[bond_idx]["length"]
            if el_tag not in out:
                out[el_tag] = {}
            out[el_tag][site_tag] = {
                "cohp": {"1": {}, "-1": {}},
                "icohp": {"-1": {}, "1": {}},
                "length": bond_length,
            }
            out[el_tag][site_tag]["cohp"]["total"] = np.zeros(len(energies))
            out[el_tag][site_tag]["icohp"]["total"] = np.zeros(len(energies))
            for spin in cohp[bond_idx]["COHP"]:
                if spin.name == "up":
                    spin_tag = "1"
                else:
                    spin_tag = "-1"
                out[el_tag][site_tag]["cohp"][spin_tag] = list(
                    cohp[bond_idx]["COHP"][spin]
                )
                out[el_tag][site_tag]["icohp"][spin_tag] = list(
                    cohp[bond_idx]["ICOHP"][spin]
                )
                out[el_tag][site_tag]["cohp"]["total"] += cohp[bond_idx]["COHP"][spin]
                out[el_tag][site_tag]["icohp"]["total"] += cohp[bond_idx]["ICOHP"][spin]

        for el_tag in out:
            for site_tag in out[el_tag]:
                for key in ["cohp", "icohp"]:
                    tmp = out[el_tag][site_tag][key]["total"]
                    out[el_tag][site_tag][key]["total"] = list(tmp)

        out["E"] = list(energies)

        write_json(out, fjson)
        return read_json(fjson)

    def tcohp(
        self, pcohp=None, fjson=None, remake=False, are_coops=False, are_cobis=False
    ):
        """
        @TODO: add demo/test

        Returns simpler dict of projected COHP, COOP, or COBI data
            - must run LOBSTER first
            - note: COHP, COOP, and COBI have similar format so they get processed the same way. If you want all three, you might do:
                for bonding in ["cohp", "coop", "cobi"]:
                    cohp_data = self.pcohp(are_coops=False, are_cobis=False)
                    coop_data = self.pcohp(are_coops=True, are_cobis=False)
                    cobi_data = self.pcohp(are_coops=False, are_cobis=True)

        Args:
            pcohp (dict) - if you already have pcohp data, you can pass it in here
            fjson (os.PathLike) - path to json file to save pdos data to
            remake (bool) - if True, remake pchohp data even if fjson exists
            are_coops (bool) - True if you want COOP analysis
            are_cobis (bool) - True if you want COBI analysis
                - both False = COHP analysis

        Returns:
            - dictionary (let's call it d)
            - list(d.keys()) = ['E'] + [el1-el2 (str) for all pairs of elements in structure that might be bonding]
                - note: el1-el2 is sorted alphabetically
                - e.g., the keys for Ca2Ru2O7 might be: ['Ca-Ca', 'Ca-Ru', 'Ca-O', 'Ru-Ru', 'O-Ru', 'O-O', 'E']
            - d['E'] = 1d array of energies corresponding with DOS
            - for each "bond type" (el1-el2), we have two keys: ['cohp', 'icohp']
                - cohp --> absolute populations
                - icohp --> integrated populations
                - d[el1-el2]['cohp' or 'icohp'] will return a 1d array of populations summed over all interactions of el1-el2 in the structure, summed over all spins

            - so if I wanted to plot the total COHP for the interaction between Ru and O throughout the structure:
                energies = d['E']
                cohp = d['O-Ru']['cohp']
                plt.plot(cohp, energies)

        """

        if are_coops:
            savename = "tcoop.json"
        elif are_cobis:
            savename = "tcobi.json"
        else:
            savename = "tcohp.json"

        if not fjson:
            fjson = os.path.join(self.calc_dir, savename)

        if os.path.exists(fjson) and not remake:
            try:
                return read_json(fjson)
            except json.decoder.JSONDecodeError:
                pass
        if not pcohp:
            pcohp = self.pcohp(are_coops=are_coops, are_cobis=are_cobis)
        if not pcohp:
            return None

        out = {"E": pcohp["E"]}
        for el_tag in pcohp:
            if el_tag == "E":
                continue
            out[el_tag] = {}
            out[el_tag]["cohp"] = np.zeros(len(out["E"]))
            out[el_tag]["icohp"] = np.zeros(len(out["E"]))
            for site_tag in pcohp[el_tag]:
                for spin in pcohp[el_tag][site_tag]["cohp"]:
                    out[el_tag]["cohp"] += np.array(
                        pcohp[el_tag][site_tag]["cohp"][spin]
                    )
                    out[el_tag]["icohp"] += np.array(
                        pcohp[el_tag][site_tag]["cohp"][spin]
                    )

        for el_tag in out:
            if el_tag == "E":
                continue
            for key in ["cohp", "icohp"]:
                tmp = out[el_tag][key]
                out[el_tag][key] = list(tmp)

        write_json(out, fjson)
        return read_json(fjson)

    def charge(self, source="bader", fjson=None, remake=False):
        """

        for obtaining partial charges resulting from dft calculations

        Args:
            fjson (_type_, optional): _description_. Defaults to None.
            remake (bool, optional): _description_. Defaults to False.

        Raises:
            NotImplementedError: _description_

        Returns:
            {el (str) : {site index (str) : {charge (float)}}}
        """
        if not fjson:
            fjson = os.path.join(self.calc_dir, "charge-%s.json" % source)
        if os.path.exists(fjson) and not remake:
            return read_json(fjson)

        if source == "bader":
            fcharge = os.path.join(self.calc_dir, "ACF.dat")
            if not os.path.exists(fcharge):
                return {}
            sites_to_els = self.sites_to_els
            nsites = len(self.outputs.contcar)
            pseudos = self.outputs.potcar
            data = {}
            with open(fcharge) as f:
                count = 0
                for line in f:
                    count += 1
                    if count < 3:
                        continue
                    if count > 2 + nsites:
                        break
                    line = [v for v in line[:-1].split(" ") if v != ""]
                    idx, charge = int(line[0]) - 1, float(line[4])
                    el = sites_to_els[idx]
                    nval = pseudos[el]["nval"]
                    delta_charge = nval - charge
                    data[idx] = {"el": el, "charge": delta_charge}
            out = {el: {} for el in list(set(sites_to_els.values()))}
            for idx in data:
                el = data[idx]["el"]
                out[el][str(idx)] = data[idx]["charge"]

        elif source in ["mulliken", "lowdin"]:
            fcharge = os.path.join(self.calc_dir, "CHARGE.lobster")
            if not os.path.exists(fcharge):
                return {}
            chg = Charge(os.path.join(self.calc_dir, "CHARGE.lobster"))
            sites_to_els = self.sites_to_els
            charges = chg.Mulliken if source == "mulliken" else chg.Loewdin
            out = {el: {} for el in list(set(sites_to_els.values()))}
            for idx in sites_to_els:
                el = sites_to_els[idx]
                out[el][str(idx)] = charges[idx]

        write_json(out, fjson)
        return read_json(fjson)

    def charged_structure(self, source, structure=False):
        """
        decorates structures with optimized magnetic moments

        Args:
            source (str): source of charge, e.g. "bader", "mulliken", "lowdin"
            structure (bool, optional): if you pass a structure, it will start from it. otherwise, it will use the CONTCAR

        Returns:
            structure where each site has a "site property" called source
            - so, if s = AnalyzeVASP(calc_dir).charged_structure(source='bader'), then s[0].properties["bader"] will give you the bader charge for that site
        """
        if not structure:
            structure = self.outputs.contcar
        charges = self.charge(source=source)
        if not charges:
            return structure
        charges_by_site = {
            idx: charge for el in charges for idx, charge in charges[el].items()
        }
        sorted_charges = [charges_by_site[str(idx)] for idx in range(len(structure))]
        structure.add_site_property(source, sorted_charges)
        return structure

    def magnetic_structure(self, structure=False):
        """
        decorates structures with optimized magnetic moments

        Args:
            structure (bool, optional): if you pass a structure, it will start from it. otherwise, it will use the CONTCAR

        Returns:
            structure where each site has a "site property" called "final_magmom"
            - so, if s = AnalyzeVASP(calc_dir).magnetic_structure(), then s[0].properties["final_magmom"] will give you the optimized magnetic moment of the first site
        """
        if not structure:
            structure = self.outputs.contcar
        mags = self.magnetization
        if not mags:
            return structure
        mags_by_site = {idx: mag for el in mags for idx, mag in mags[el].items()}
        sorted_mags = [mags_by_site[str(idx)] for idx in range(len(structure))]
        structure.add_site_property("final_magmom", sorted_mags)
        return structure

    def decorated_structure(self, charge_source="bader", structure=None):
        """
        decorates structures with optimized magnetic moments

        Args:
            structure (bool, optional): if you pass a structure, it will start from it. otherwise, it will use the CONTCAR

        Returns:
            structure where each site has two "site properties": ["final_magmom", charge_source]
            - so, if s = AnalyzeVASP(calc_dir).magnetic_structure(), then s[0].properties["final_magmom"] will give you the optimized magnetic moment of the first site
        """
        if not structure:
            structure = self.outputs.contcar
        mag_struc = self.magnetic_structure(structure=structure)
        dec_struc = self.charged_structure(source=charge_source, structure=mag_struc)
        return dec_struc

    @property
    def E_madelung(self):
        """
        Returns the Madelung energy for a structure
            - must run lobster

        Returns:
            {'mulliken' : Madelung energy using Mulliken charges,
            'lowdin' : Madelung energy using Lowdin charges}
        """
        fmadelung = os.path.join(self.calc_dir, "MadelungEnergies.lobster")
        if os.path.exists(fmadelung):
            return {
                "mulliken": MadelungEnergies(
                    filename=fmadelung
                ).madelungenergies_Mulliken,
                "lowdin": MadelungEnergies(filename=fmadelung).madelungenergies_Loewdin,
            }
        return {}

    @property
    def formula(self):
        """
        returns the compact formula of the calculated structure

        Returns:
            compact formula (str) or None
        """
        contcar = self.outputs.contcar
        if contcar:
            return StrucTools(contcar).compact_formula
        poscar = self.outputs.poscar
        if poscar:
            return StrucTools(poscar).compact_formula
        return None

    @property
    def spacegroup_info(self):
        """
        Returns:
            dict of spacegroup info with 'tight' or 'loose' symmetry tolerance
            e.g.,
                data['tight']['number'] returns spacegroup number with tight tolerance
                data['loose']['symbol'] returns spacegroup symbol with loose tolerance

        """
        if self.outputs.contcar:
            return StrucTools(self.outputs.contcar).spacegroup_info
        else:
            return None

    @property
    def basic_info(self):
        """
        Returns:
            {'convergence' : True if calc converged else False,
            'E_per_at' : energy per atom if calc converged else None,
            'formula' : compact formula of computed structure,
            'sg' : dict of spacegroup information}
        """
        E_per_at = self.E_per_at
        convergence = True if E_per_at else False
        return {
            "convergence": convergence,
            "E_per_at": E_per_at,
            "formula": self.formula,
            "sg": self.spacegroup_info,
        }

    @property
    def relaxed_structure(self):
        """

        Returns:
            pymatgen Structure.as_dict() for relaxed structure
        """
        structure = self.outputs.contcar
        if structure:
            return structure.as_dict()
        else:
            return None

    @property
    def metadata(self):
        """returns compiled metadata

        Returns:
            {'incar' : incar.as_dict(),
             'all_input_settings' : all_input_settings.as_dict(),
             'kpoints' : kpoints.as_dict(),
             'potcar' : potcar.as_dict(),
             'calc_dir' : calc_dir,}
        """
        outputs = self.outputs
        meta = {}
        incar_data = outputs.incar
        if not incar_data:
            meta["incar"] = {}
        else:
            meta["incar"] = (
                incar_data if isinstance(incar_data, dict) else incar_data.as_dict()
            )

        kpoints_data = outputs.kpoints
        if not kpoints_data:
            meta["kpoints"] = {}
        else:
            meta["kpoints"] = (
                kpoints_data
                if isinstance(kpoints_data, dict)
                else kpoints_data.as_dict()
            )

        potcar_data = outputs.potcar
        if not potcar_data:
            meta["potcar"] = {}
        else:
            meta["potcar"] = potcar_data

        all_input_settings = outputs.all_input_settings
        meta["all_input_settings"] = (
            all_input_settings
            if isinstance(all_input_settings, dict)
            else all_input_settings.as_dict()
        )

        meta["calc_dir"] = self.calc_dir

        return meta

    @property
    def calc_setup(self):
        """

        Returns:
            self explanatory?
        """
        calc_dir = self.calc_dir
        formula, ID, standard, mag, xc_calc = calc_dir.split("/")[-5:]
        return {
            "formula_tag": formula,
            "ID": ID,
            "standard": standard,
            "mag": mag,
            "xc": xc_calc.split("-")[0],
            "calc": xc_calc.split("-")[1],
        }

    def summary(
        self,
        include_meta=False,
        include_calc_setup=False,
        include_structure=False,
        include_mag=False,
        include_tdos=False,
        include_pdos=False,
        include_gap=True,
        include_charge=True,
        include_madelung=True,
        include_tcohp=False,
        include_pcohp=False,
        include_tcoop=False,
        include_pcoop=False,
        include_tcobi=False,
        include_pcobi=False,
    ):
        """
        Returns all desired data for post-processing DFT calculations

        Args:
            include_meta (bool, optional): _description_. Defaults to False.
            include_calc_setup (bool, optional): _description_. Defaults to False.
            include_structure (bool, optional): _description_. Defaults to False.
            include_mag (bool, optional): _description_. Defaults to False.
            include_dos (bool, optional): _description_. Defaults to False.
            include_gap (bool, optional): _description_. Defaults to True.

        Raises:
            NotImplementedError: _description_

        Returns:
            _type_: _description_
        """
        data = {}
        data["results"] = self.basic_info
        if include_meta:
            data["meta"] = self.metadata
        if include_calc_setup:
            if "meta" not in data:
                data["meta"] = {}
            data["meta"]["setup"] = self.calc_setup
        if include_structure:
            # data["structure"] = self.decorated_structure(
            #    charge_source="bader", structure=None
            # )
            data["structure"] = self.relaxed_structure
        if include_mag:
            data["magnetization"] = self.magnetization
        if include_tdos:
            pdos = self.pdos()
            tdos = self.tdos(pdos=pdos)
            data["tdos"] = tdos
        if include_pdos:
            pdos = self.pdos()
            data["pdos"] = pdos
        if include_gap:
            data["gap"] = self.gap_properties
        if include_charge:
            data["charge"] = {}
            for source in ["bader", "mulliken", "lowdin"]:
                data["charge"][source] = self.charge(source)
        if include_madelung:
            data["madelung"] = self.E_madelung
        if include_tcohp:
            pcohp = self.pcohp()
            tcohp = self.tcohp(pcohp=pcohp)
            data["tcohp"] = tcohp
        if include_pcohp:
            pcohp = self.pcohp()
            data["pcohp"] = pcohp
        if include_tcoop:
            data["tcoop"] = self.tcohp(are_coops=True)
        if include_pcoop:
            data["pcoop"] = self.pcohp(are_coops=True)
        if include_tcobi:
            data["tcobi"] = self.tcohp(are_cobis=True)
        if include_pcobi:
            data["pcobi"] = self.pcohp(are_cobis=True)
        return data


class AnalyzeBatch(object):
    def __init__(
        self,
        launch_dirs,
        user_configs={},
        analysis_configs_yaml=os.path.join(
            os.getcwd(), "_batch_VASP_analysis_configs.yaml"
        ),
        refresh_configs=True,
    ):
        """
        Args:
            launch_dirs (dict) : {launch directory : {'xcs' : [final_xcs for each chain of jobs], 'magmom' : [list of magmoms for that launch directory]}}
            user_configs (dict) : any configs relevant to analysis
                only_static: True # only retrieve data from the static calculations
                one_xc: if None, retrieve all xcs, else retrieve only the one specified
                check_relax: True # make sure the relax calculation and the static have similar energies
                include_meta: False # include metadata like INCAR, KPOINTS, POTCAR settings
                include_calc_setup: True # include things related to the calculation setup -- standard, mag, final_xc, etc
                include_structure: True # include the relaxed crystal structure as a dict
                include_mag: False # include the relaxed magnetization info as as dict
                include_dos: False # include the density of states
                verbose: True # print stuff as things get analyzed
                n_procs: all # how many cores to parallelize the analysis over
            analysis_configs_yaml (str) : path to yaml file with baseline analysis configs
            refresh_configs (bool): if True, will refresh the local baseline analysis configs with the pydmc version
        """

        # should get these from LaunchTools
        self.launch_dirs = launch_dirs

        # write baseline analysis configs locally if not there or want to be refreshed
        if not os.path.exists(analysis_configs_yaml) or refresh_configs:
            _analysis_configs = load_batch_vasp_analysis_configs()
            write_yaml(_analysis_configs, analysis_configs_yaml)

        _analysis_configs = read_yaml(analysis_configs_yaml)

        # update configs with any user_configs
        configs = {**_analysis_configs, **user_configs}

        # figure out how many processors to use
        if configs["n_procs"] == "all":
            configs["n_procs"] = multip.cpu_count() - 1

        # copy configs to prevent unwanted changes
        self.configs = configs.copy()

    @property
    def calc_dirs(self):
        """
        Returns:
            a list of all calculation directories to crawl through and collect VASP output info
        """
        launch_dirs = self.launch_dirs
        all_calc_dirs = []
        calcs = (
            ["loose", "relax", "static"]
            if not self.configs["only_static"]
            else ["static"]
        )
        for launch_dir in launch_dirs:
            files_in_launch_dir = os.listdir(launch_dir)
            calc_dirs = [
                os.path.join(launch_dir, c)
                for c in files_in_launch_dir
                if "-" in c
                if c.split("-")[1] in calcs
            ]
            calc_dirs = [
                c for c in calc_dirs if os.path.exists(os.path.join(c, "POSCAR"))
            ]
            all_calc_dirs += calc_dirs
        return sorted(list(set(all_calc_dirs)))

    def _key_for_calc_dir(self, calc_dir):
        """
        Args:
            calc_dir (str) : path to a calculation directory where VASP was executed

        Returns:
            a string that can be used as a key for a dictionary
                top_level.unique_ID.standard.mag.xc_calc
        """
        return ".".join(calc_dir.split("/")[-5:])

    def _results_for_calc_dir(self, calc_dir):
        """
        Args:
            calc_dir (str) : path to a calculation directory where VASP was executed

        Returns:
            a dictionary of results for that calculation directory
                - format varies based on self.configs
                - see AnalyzeVASP.summary() for more info
        """

        configs = self.configs.copy()
        verbose = configs["verbose"]
        include_meta = configs["include_meta"]
        include_calc_setup = configs["include_calc_setup"]
        include_structure = configs["include_structure"]
        include_mag = configs["include_mag"]
        include_tdos = configs["include_tdos"]
        include_pdos = configs["include_pdos"]
        include_tcohp = configs["include_tcohp"]
        include_pcohp = configs["include_pcohp"]
        include_tcoop = configs["include_tcoop"]
        include_pcoop = configs["include_pcoop"]
        include_tcobi = configs["include_tcobi"]
        include_pcobi = configs["include_pcobi"]
        check_relax = configs["check_relax"]
        create_cif = configs["create_cif"]

        if verbose:
            print("analyzing %s" % calc_dir)
        analyzer = AnalyzeVASP(calc_dir)

        # collect the data we asked for
        summary = analyzer.summary(
            include_meta=include_meta,
            include_calc_setup=include_calc_setup,
            include_structure=include_structure,
            include_mag=include_mag,
            include_tdos=include_tdos,
            include_pdos=include_pdos,
            include_tcohp=include_tcohp,
            include_pcohp=include_pcohp,
            include_tcoop=include_tcoop,
            include_pcoop=include_pcoop,
            include_tcobi=include_tcobi,
            include_pcobi=include_pcobi,
        )

        # store the relax energy if we asked to
        if check_relax:
            relax_energy = AnalyzeVASP(calc_dir.replace("static", "relax")).E_per_at
            summary["meta"]["E_relax"] = relax_energy
            if not relax_energy:
                summary["results"]["convergence"] = False
                summary["results"]["E_per_at"] = None

        # save the data in a dictionary with a key for that calc_dir
        key = self._key_for_calc_dir(calc_dir)

        if create_cif and summary["results"]["convergence"] and include_structure:
            if summary["structure"]:
                s = Structure.from_dict(summary["structure"])
                s.to(fmt="cif", filename=os.path.join(calc_dir, key + ".cif"))
            else:
                print("no structure, cant make cif")
        return {key: summary}

    @property
    def results(self):
        """

        Returns:
            {calc_dir key : results for that calc_dir}
        """

        n_procs = self.configs["n_procs"]

        calc_dirs = self.calc_dirs

        # run serial if only one processor
        if n_procs == 1:
            data = [self._results_for_calc_dir(calc_dir) for calc_dir in calc_dirs]

        # otherwise, run parallel
        if n_procs > 1:
            pool = multip.Pool(processes=n_procs)
            data = pool.map(self._results_for_calc_dir, calc_dirs)
            pool.close()

        # each item in data is a dictionary that looks like {key : data for that key}
        # we'll map to a dictionary of {key : data for that key} for all keys
        out = {}
        for d in data:
            for key in d:
                out[key] = d[key]

        return out


def crawl_and_purge(
    head_dir,
    files_to_purge=[
        "WAVECAR",
        "CHGCAR",
        "CHG",
        "PROCAR",
        "LOCPOT",
        "AECCAR0",
        "AECCAR1",
        "AECCAR2",
    ],
    safety="on",
    check_convergence=True,
    verbose=False,
):
    """
    Args:
        head_dir (str) - directory to start crawling beneath
        files_to_purge (list) - list of file names to purge
        safety (str) - 'on' or 'off' to turn on/off safety
            - if safety is on, won't actually delete files
    """
    purged_files = []
    mem_created = 0
    for subdir, dirs, files in os.walk(head_dir):
        ready = False
        if check_convergence:
            if "POTCAR" in files:
                av = AnalyzeVASP(subdir)
                if av.is_converged:
                    ready = True
                else:
                    ready = False
            else:
                ready = False
        else:
            ready = True
        if ready:
            for f in files:
                if f in files_to_purge:
                    path_to_f = os.path.join(subdir, f)
                    if verbose:
                        print(path_to_f)
                    mem_created += os.stat(path_to_f).st_size
                    purged_files.append(path_to_f)
                    if safety == "off":
                        os.remove(path_to_f)
    if safety == "off":
        print(
            "You purged %i files, freeing up %.2f GB of memory"
            % (len(purged_files), mem_created / 1e9)
        )
    if safety == "on":
        print(
            "You had the safety on\n If it were off, you would have purged %i files, freeing up %.2f GB of memory"
            % (len(purged_files), mem_created / 1e9)
        )



from pydmclab.core.struc import StrucTools, SiteTools
from pydmclab.core.query import MPQuery

from pymatgen.core.structure import Structure

MPID = "mp-19417"
API_KEY = "N3KdATtMmcsUL94g"


def demo_basic_structure_manipulations(mpid=MPID):
    print("\n")

    s = MPQuery(API_KEY).get_structure_by_material_id(mpid)
    st = StrucTools(s)

    print("\nformula = %s" % st.formula)
    print("compact formula = %s" % st.compact_formula)
    print("structure has %s" % st.els)
    print("downloaded structure has %i sites" % len(st.structure))
    st.make_supercell([1, 2, 3])
    print("supercell has %i sites" % len(st.structure))


def demo_ox_state_decoration(mpid=MPID):
    print("\n")
    s = MPQuery(API_KEY).get_structure_by_material_id(mpid)
    st = StrucTools(s)

    site = SiteTools(s, 0)
    print("from MP, site 0 is %s" % site.ion)

    s = st.decorate_with_ox_states

    print("after auto oxi state decoration, site 10 is %s" % SiteTools(s, 10).ion)

    st = StrucTools(s, ox_states={"Fe": 2, "Ti": 4, "O": -2})

    s = st.decorate_with_ox_states

    print(
        "after forcing Fe to be ox state = 2 (bad idea)!, site 10 is %s"
        % SiteTools(s, 10).ion
    )

    print(s[0].species)


def demo_replace_species_and_order(mpid=MPID):
    print("\n")
    s = MPQuery(API_KEY).get_structure_by_material_id(mpid)
    st = StrucTools(s)
    st.make_supercell([1, 2, 3])

    species_map = {"Fe": {"Fe": 0.875, "Cr": 0.125}}

    ordered_strucs = st.replace_species(species_map, n_strucs=10)

    print("generated %i structures" % len(ordered_strucs))

    for idx in ordered_strucs:
        s = Structure.from_dict(ordered_strucs[idx])
        print("\nstructure %i has formula %s" % (idx, StrucTools(s).formula))
        site = SiteTools(s, 12)
        print(
            "site 13 has ion = %s, coords = %s with ox state = %s"
            % (site.ion, site.coords, site.ox_state)
        )


def demo_dilute_vacancy(mpid=MPID):
    """
    Broken right now...
    """
    print("\n")
    s = MPQuery(API_KEY).get_structure_by_material_id(mpid)
    st = StrucTools(s)

    out = st.get_structures_with_dilute_vacancy(
        el_to_replace="O", n_strucs=10, structure=None
    )

    struc = Structure.from_dict(out[0])

    print(struc)


def main():
    demo_basic_structure_manipulations()
    demo_ox_state_decoration(mpid=MPID)
    demo_replace_species_and_order(mpid=MPID)
    # demo_dilute_vacancy(mpid=MPID)
    return


if __name__ == "__main__":
    main()



import matplotlib as mpl
import os
import numpy as np
from pydmclab.utils.handy import read_json, write_json

this_dir, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(this_dir, "data")


def get_colors_dict(palette, names):

    palette = mpl.cm.get_cmap(palette)
    # Create an empty dictionary to store the mappings
    color_dict = {}

    # Loop through each color in the colormap

    for i in range(palette.N):
        # Convert the color index to a RGB tuple
        rgb = tuple(np.array(palette(i)[:3]))
        # Add the RGB tuple to the dictionary, using the color name as the key
        color_dict[names[i]] = rgb

    return color_dict


def get_color_palettes(remake=False):
    fjson = os.path.join(DATA_PATH, "colors.json")
    if not remake and os.path.exists(fjson):
        return read_json(fjson)
    tab10 = mpl.cm.get_cmap("tab10")
    tab10_names = [
        "blue",
        "orange",
        "green",
        "red",
        "purple",
        "brown",
        "pink",
        "gray",
        "yellow",
        "cyan",
    ]

    paired = mpl.cm.get_cmap("Paired")
    paired_names = [
        "light blue",
        "blue",
        "light green",
        "green",
        "light red",
        "red",
        "light orange",
        "orange",
        "light purple",
        "purple",
        "light brown",
        "brown",
    ]

    set2 = mpl.cm.get_cmap("Set2")
    set2_names = [
        "green",
        "orange",
        "blue",
        "pink",
        "light green",
        "yellow",
        "brown",
        "gray",
    ]

    dark2 = mpl.cm.get_cmap("Dark2")
    dark2_names = [
        "green",
        "orange",
        "purple",
        "pink",
        "light green",
        "yellow",
        "brown",
        "gray",
    ]

    palettes = [tab10, paired, set2, dark2]
    palette_names = ["tab10", "paired", "set2", "dark2"]
    color_names = [tab10_names, paired_names, set2_names, dark2_names]

    data = {}
    for i, palette in enumerate(palettes):
        data[palette_names[i]] = get_colors_dict(palette, color_names[i])

    write_json(data, fjson)
    return read_json(fjson)


def main():
    colors = get_color_palettes()
    return colors


if __name__ == "__main__":
    d = main()
